/*
 * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
 * Copyright (C) 1999-2001 Hiroyuki Yamamoto
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "defs.h"

#include <glib.h>
#include <string.h>

#include "intl.h"
#include "main.h"
#include "address.h"
#include "prefs.h"
#include "xml.h"
#include "procmime.h"
#include "utils.h"
#include "codeconv.h"

static AddressTree address_tree;

static void address_build_tree			(XMLFile	*file,
						 GNode		*node,
						 const gchar	*folder_tag);
static void address_add_objects			(XMLFile	*file,
						 GNode		*node);

static AddressItem *address_parse_item		(XMLFile	*file);
static void address_xml_recursive_write		(GNode		*node,
						 FILE		*fp);
static void address_node_write_begin		(GNode		*node,
						 FILE		*fp);
static void address_node_write_end		(GNode		*node,
						 FILE		*fp);
static void address_write_items			(FILE		*fp,
						 GList		*items,
						 guint		 level);
static void tab_indent_out			(FILE		*fp,
						 guint		 level);


AddressItem *address_parse_str(const gchar *str)
{
	gchar *name    = NULL;
	gchar *address = NULL;
	gchar *buf;
	gchar *start, *end;

	Xalloca(buf, strlen(str) + 1, return NULL);

	strcpy(buf, str);
	g_strstrip(buf);
	if (*buf == '\0') return NULL;

	if ((start = strchr(buf, '<'))) {
		if (start > buf) {
			*start = '\0';
			g_strstrip(buf);
			if (*buf != '\0')
				name = buf;
		}
		start++;
		if ((end = strchr(start, '>'))) {
			*end = '\0';
			g_strstrip(start);
			if (*start != '\0')
				address = start;
		}
	} else
		name = buf;

	if (!name && !address) return NULL;

	return ADDRESS_ITEM(address_item_new(name, address, NULL));
}

AddressObject *address_folder_new(const gchar *name)
{
	AddressFolder *folder;

	folder = g_new(AddressFolder, 1);
	ADDRESS_OBJECT_TYPE(folder) = ADDR_FOLDER;
	ADDRESS_OBJECT(folder)->node = NULL;
	folder->name = g_strdup(name);
	folder->items = NULL;

	return ADDRESS_OBJECT(folder);
}

AddressObject *address_group_new(const gchar *name)
{
	AddressGroup *group;

	group = g_new(AddressGroup, 1);
	ADDRESS_OBJECT_TYPE(group) = ADDR_GROUP;
	ADDRESS_OBJECT(group)->node = NULL;
	group->name = g_strdup(name);
	group->items = NULL;

	return ADDRESS_OBJECT(group);
}

AddressObject *address_item_new(const gchar *name, const gchar *address,
				const gchar *remarks)
{
	AddressItem *item;

	item = g_new(AddressItem, 1);
	ADDRESS_OBJECT_TYPE(item) = ADDR_ITEM;
	ADDRESS_OBJECT(item)->node = NULL;
	item->name    = g_strdup(name);
	item->address = g_strdup(address);
	item->remarks = g_strdup(remarks);

	return ADDRESS_OBJECT(item);
}

void address_item_destroy(AddressItem *item)
{
	g_return_if_fail(item != NULL);

	g_free(item->name);
	g_free(item->address);
	g_free(item->remarks);
	g_free(item);
}

gchar *address_get_name(AddressObject *obj)
{
	g_return_val_if_fail(obj != NULL, NULL);

	if (obj->type == ADDR_GROUP)
		return ADDRESS_GROUP(obj)->name;
	else if (obj->type == ADDR_FOLDER)
		return ADDRESS_FOLDER(obj)->name;
	else if (obj->type == ADDR_ITEM)
		return ADDRESS_ITEM(obj)->name;

	return NULL;
}

GList *address_get_items(AddressObject *obj)
{
	g_return_val_if_fail(obj != NULL, NULL);

	if (obj->type == ADDR_GROUP)
		return ADDRESS_GROUP(obj)->items;
	else if (obj->type == ADDR_FOLDER)
		return ADDRESS_FOLDER(obj)->items;
	else
		return NULL;
}

GNode *address_get_tree(void)
{
	return address_tree.node;
}

static gboolean address_get_folder_list_func(GNode *node, gpointer data)
{
	GList **list = (GList **)data;
	AddressObject *obj = ADDRESS_OBJECT(node->data);

	if (!node->parent) return FALSE;

	if (ADDRESS_OBJECT_TYPE(obj) == ADDR_FOLDER ||
	    ADDRESS_OBJECT_TYPE(obj) == ADDR_GROUP)
		*list = g_list_append(*list, obj);

	return FALSE;
}

GList *address_get_folder_list(void)
{
	GList *list = NULL;

	g_node_traverse(address_tree.node, G_PRE_ORDER, G_TRAVERSE_ALL, -1,
			address_get_folder_list_func, &list);

	return list;
}

void address_read_file(void)
{
	XMLFile *file;
	gchar *path;
	AddressObject *root;
	AddressObject *common_folder;
	AddressObject *personal_folder;

	debug_print(_("Reading addressbook file..."));

	root = address_folder_new(NULL);
	root->node = address_tree.node = g_node_new(root);
	common_folder = address_folder_new(ADDRESS_TAG_COMMON);
	personal_folder = address_folder_new(ADDRESS_TAG_PERSONAL);
	address_add_object(root, common_folder);
	address_add_object(root, personal_folder);
	address_tree.common   = common_folder;
	address_tree.personal = personal_folder;

	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, ADDRESS_BOOK, NULL);
	if ((file = xml_open_file(path)) == NULL) {
		debug_print(_("%s doesn't exist.\n"), path);
		g_free(path);
		address_build_tree(NULL, common_folder->node, ADDRESS_TAG_COMMON);
		address_build_tree(NULL, personal_folder->node, ADDRESS_TAG_PERSONAL);
		return;
	}
	g_free(path);

	xml_get_dtd(file);

	if (xml_parse_next_tag(file) < 0 ||
	    xml_compare_tag(file, "addressbook") == FALSE) {
		g_warning("Invalid addressbook data\n");
		xml_close_file(file);
		return;
	}

	address_build_tree(file, common_folder->node, ADDRESS_TAG_COMMON);
	address_build_tree(file, personal_folder->node, ADDRESS_TAG_PERSONAL);

	xml_close_file(file);

	debug_print(_("done.\n"));
}

static void address_build_tree(XMLFile *file, GNode *node,
			       const gchar *folder_tag)
{
	g_return_if_fail(node != NULL);
	g_return_if_fail(folder_tag != NULL);

	if (!file) return;

	if (xml_parse_next_tag(file) < 0 ||
	    xml_compare_tag(file, folder_tag) == FALSE) {
		g_warning("Invalid addressbook data\n");
		return;
	}

	address_add_objects(file, node);
}

static void address_add_objects(XMLFile *file, GNode *node)
{
	GList *attr;
	guint prev_level;
	AddressObject *obj = ADDRESS_OBJECT(node->data);

	for (;;) {
		prev_level = file->level;
		if (xml_parse_next_tag(file) < 0) return;
		if (file->level < prev_level) return;

		if (xml_compare_tag(file, "group")) {
			AddressObject *group;

			attr = xml_get_current_tag_attr(file);
			group = address_group_new
				(attr ? ((XMLAttr *)attr->data)->value : NULL);

			address_add_object(obj, group);
			address_add_objects(file, group->node);
		} else if (xml_compare_tag(file, "folder")) {
			AddressObject *folder;

			attr = xml_get_current_tag_attr(file);
			folder = address_folder_new
				(attr ? ((XMLAttr *)attr->data)->value : NULL);

			address_add_object(obj, folder);
			address_add_objects(file, folder->node);
		} else if (xml_compare_tag(file, "item")) {
			AddressItem *item;

			item = address_parse_item(file);
			if (!item) return;
			address_add_object(obj, ADDRESS_OBJECT(item));
		} else {
			g_warning("Invalid tag\n");
			return;
		}
	}
}

void address_add_object(AddressObject *parent, AddressObject *obj)
{
	g_return_if_fail(parent != NULL);
	g_return_if_fail(obj    != NULL);
	g_return_if_fail(parent->node != NULL);

	if (parent->type == ADDR_ITEM) {
		g_warning("Parent object mustn't be an item.\n");
		return;
	}

	if (obj->type == ADDR_GROUP) {
		if (parent->type != ADDR_FOLDER) {
			g_warning("Group can't be added in another group.\n");
			return;
		}

		obj->node = g_node_append_data(parent->node, obj);
	} else if (obj->type == ADDR_FOLDER) {
		if (parent->type != ADDR_FOLDER) {
			g_warning("Group can't contain folders.\n");
			return;
		}

		obj->node = g_node_append_data(parent->node, obj);
	}

	if (obj->type == ADDR_GROUP || obj->type == ADDR_ITEM) {
		if (parent->type == ADDR_GROUP) {
			AddressGroup *group = ADDRESS_GROUP(parent);
			group->items = g_list_append(group->items, obj);
		} else if (parent->type == ADDR_FOLDER) {
			AddressFolder *folder = ADDRESS_FOLDER(parent);
			folder->items = g_list_append(folder->items, obj);
		}
	}

	//gtk_ctree_sort_node(ctree, parent->node);
}

static gboolean address_delete_object_func(GNode *node, gpointer data)
{
	AddressObject *obj = ADDRESS_OBJECT(node->data);

	g_return_val_if_fail(ADDRESS_OBJECT_TYPE(obj) == ADDR_GROUP ||
			     ADDRESS_OBJECT_TYPE(obj) == ADDR_FOLDER,
			     FALSE);

	if (ADDRESS_OBJECT_TYPE(obj) == ADDR_GROUP) {
		AddressGroup *group = ADDRESS_GROUP(obj);

		g_free(group->name);
		while (group->items != NULL) {
			obj = ADDRESS_OBJECT(group->items->data);
			if (ADDRESS_OBJECT_TYPE(obj) == ADDR_ITEM)
				address_item_destroy(ADDRESS_ITEM(obj));
			else
				g_warning("Group shouldn't contain other than item.\n");
			group->items = g_list_remove(group->items, obj);
		}
		g_free(group);
	} else if (ADDRESS_OBJECT_TYPE(obj) == ADDR_FOLDER) {
		AddressFolder *folder = ADDRESS_FOLDER(obj);

		g_free(folder->name);
		while (folder->items != NULL) {
			obj = ADDRESS_OBJECT(folder->items->data);
			if (ADDRESS_OBJECT_TYPE(obj) == ADDR_ITEM)
				address_item_destroy(ADDRESS_ITEM(obj));
			folder->items = g_list_remove(folder->items, obj);
		}
		g_free(folder);
	}

	return FALSE;
}

void address_delete_object(AddressObject *obj)
{
	g_return_if_fail(obj != NULL);

	if (ADDRESS_OBJECT_TYPE(obj) == ADDR_ITEM)
		address_item_destroy(ADDRESS_ITEM(obj));
	else {
		GNode *node = obj->node;
		g_node_traverse(node, G_PRE_ORDER, G_TRAVERSE_ALL, -1,
				address_delete_object_func, NULL);
		g_node_destroy(node);
	}
}

AddressObject *address_find_object_by_name(AddressObject *obj,
					   const gchar *name)
{
	GList *found;

	g_return_val_if_fail(obj != NULL, NULL);

	if (obj->type == ADDR_GROUP) {
		AddressGroup *group = ADDRESS_GROUP(obj);

		found = g_list_find_custom(group->items, (gpointer)name,
					   address_obj_name_compare);
		if (found) return ADDRESS_OBJECT(found->data);
	} else if (obj->type == ADDR_FOLDER) {
		AddressFolder *folder = ADDRESS_FOLDER(obj);

		found = g_list_find_custom(folder->items, (gpointer)name,
					   address_obj_name_compare);
		if (found) return ADDRESS_OBJECT(found->data);
	} else if (obj->type == ADDR_ITEM) {
		if (!address_obj_name_compare(obj, name)) return obj;
	}

	return NULL;
}

AddressObject *address_find_item_by_name_and_address(AddressObject *obj,
						     const gchar *name,
						     const gchar *address)
{
	AddressObject *found = NULL;
	AddressObject *item;
	GList *list;

	g_return_val_if_fail(obj != NULL, NULL);

	item = address_item_new(name, address, NULL);

	if (ADDRESS_OBJECT_TYPE(obj) == ADDR_GROUP) {
		AddressGroup *group = ADDRESS_GROUP(obj);

		list = g_list_find_custom(group->items, item,
					  address_item_compare);
		if (list) found = ADDRESS_OBJECT(list->data);
	} else if (ADDRESS_OBJECT_TYPE(obj) == ADDR_FOLDER) {
		AddressFolder *folder = ADDRESS_FOLDER(obj);

		list = g_list_find_custom(folder->items, item,
					  address_item_compare);
		if (list) found = ADDRESS_OBJECT(list->data);
	} else if (ADDRESS_OBJECT_TYPE(obj) == ADDR_ITEM) {
		if (!address_item_compare(obj, item)) found = obj;
	}

	address_item_destroy(ADDRESS_ITEM(item));
	return found;
}

static AddressItem *address_parse_item(XMLFile *file)
{
	gchar *element;
	AddressItem *item;
	guint level;

	item = g_new0(AddressItem, 1);
	ADDRESS_OBJECT(item)->type = ADDR_ITEM;

	level = file->level;

	while (xml_parse_next_tag(file) == 0) {
		if (file->level < level) return item;
		if (file->level == level) break;

		element = xml_get_element(file);

		if (xml_compare_tag(file, "name")) {
			item->name = element;
		} else if (xml_compare_tag(file, "address")) {
			item->address = element;
		} else if (xml_compare_tag(file, "remarks")) {
			item->remarks = element;
		}

		if (xml_parse_next_tag(file) < 0) break;
		if (file->level != level) break;
	}

	g_warning("address_parse_item(): Parse error\n");
	address_item_destroy(item);
	return NULL;
}

void address_export_to_file(void)
{
	PrefFile *pfile;
	gchar *path;

	if (!address_tree.node) return;

	debug_print(_("Exporting addressbook to file..."));

	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, ADDRESS_BOOK, NULL);
	if ((pfile = prefs_write_open(path)) == NULL) {
		g_free(path);
		return;
	}
	g_free(path);

	fprintf(pfile->fp, "<?xml version=\"1.0\" encoding=\"%s\"?>\n\n",
		conv_get_current_charset_str());

	address_xml_recursive_write(NULL, pfile->fp);

	if (prefs_write_close(pfile) < 0) {
		g_warning(_("failed to write addressbook data.\n"));
		return;
	}

	debug_print(_("done.\n"));
}

static void address_xml_recursive_write(GNode *node, FILE *fp)
{
	GNode *work;
	GNode *tmp;

	if (node) {
		work = node->children;
		address_node_write_begin(node, fp);
	} else
		work = address_tree.node;

	while (work) {
		tmp = work->next;
		address_xml_recursive_write(work, fp);
		work = tmp;
	}

	if (node)
		address_node_write_end(node, fp);
}

static void address_node_write_begin(GNode *node, FILE *fp)
{
	AddressObject *obj;
	guint depth;

	obj = ADDRESS_OBJECT(node->data);
	g_return_if_fail(obj != NULL);

	depth = g_node_depth(node);

	if (obj->type == ADDR_FOLDER) {
		AddressFolder *folder = ADDRESS_FOLDER(obj);

		tab_indent_out(fp, depth - 1);
		if (depth == 1)
			fputs("<addressbook>\n", fp);
		else if (depth == 2)
			fprintf(fp, "<%s>\n", folder->name);
		else {
			fputs("<folder name=\"", fp);
			xml_file_put_escape_str(fp, folder->name);
			fputs("\">\n", fp);
		}
	} else if (obj->type == ADDR_GROUP) {
		AddressGroup *group = ADDRESS_GROUP(obj);

		tab_indent_out(fp, depth - 1);
		fputs("<group name=\"", fp);
		xml_file_put_escape_str(fp, group->name);
		fputs("\">\n", fp);
	}
}

static void address_node_write_end(GNode *node, FILE *fp)
{
	AddressObject *obj;
	guint depth;

	obj = ADDRESS_OBJECT(node->data);
	g_return_if_fail(obj != NULL);

	depth = g_node_depth(node);

	if (obj->type == ADDR_FOLDER) {
		AddressFolder *folder = ADDRESS_FOLDER(obj);

		address_write_items(fp, folder->items, depth);

		tab_indent_out(fp, depth - 1);
		if (depth == 1)
			fputs("</addressbook>\n", fp);
		else if (depth == 2)
			fprintf(fp, "</%s>\n", folder->name);
		else
			fputs("</folder>\n", fp);
	} else if (obj->type == ADDR_GROUP) {
		AddressGroup *group = ADDRESS_GROUP(obj);

		address_write_items(fp, group->items, depth);

		tab_indent_out(fp, depth - 1);
		fputs("</group>\n", fp);
	}
}

static void address_write_items(FILE *fp, GList *items, guint level)
{
	AddressItem *item;

	for (; items != NULL; items = items->next) {
		if (ADDRESS_OBJECT_TYPE(items->data) == ADDR_ITEM) {
			item = ADDRESS_ITEM(items->data);

			tab_indent_out(fp, level);
			fputs("<item>\n", fp);

			tab_indent_out(fp, level + 1);
			fputs("<name>", fp);
			xml_file_put_escape_str(fp, item->name);
			fputs("</name>\n", fp);

			tab_indent_out(fp, level + 1);
			fputs("<address>", fp);
			xml_file_put_escape_str(fp, item->address);
			fputs("</address>\n", fp);

			tab_indent_out(fp, level + 1);
			fputs("<remarks>", fp);
			xml_file_put_escape_str(fp, item->remarks);
			fputs("</remarks>\n", fp);

			tab_indent_out(fp, level);
			fputs("</item>\n", fp);
		}
	}
}

static void tab_indent_out(FILE *fp, guint level)
{
	guint i;

	for (i = 0; i < level; i++)
		fputs("    ", fp);
}

gint address_obj_name_compare(gconstpointer a, gconstpointer b)
{
	const AddressObject *obj = a;
	const gchar *name = b;
	const gchar *obj_name;

	if (!obj || !name) return -1;

	obj_name = address_get_name(ADDRESS_OBJECT(obj));
	if (obj_name)
		return g_strcasecmp(obj_name, name);
	else
		return -1;
}

gint address_item_compare(gconstpointer a, gconstpointer b)
{
	const AddressObject *obj = a;
	const AddressObject *item = b;
	const gchar *name;
	const gchar *address;
	const gchar *obj_name;
	const gchar *obj_address;

	if (!obj || !item) return -1;
	if (ADDRESS_OBJECT_TYPE(obj)  != ADDR_ITEM) return -1;
	if (ADDRESS_OBJECT_TYPE(item) != ADDR_ITEM) return -1;

	obj_name = ADDRESS_ITEM(obj)->name;
	obj_address = ADDRESS_ITEM(obj)->address;
	name = ADDRESS_ITEM(item)->name;
	address = ADDRESS_ITEM(item)->address;

	if (obj_name && name && strcmp(obj_name, name) != 0) return -1;
	if (!obj_address || !address) return -1;
	return g_strcasecmp(obj_address, address);
}
